# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 09:13 (EST)
# Function: config file reading
#
# $Id: Conf.pm,v 1.18 2002/10/28 23:15:13 jaw Exp $

# some bozo put a Config.pm in the standard perl dist
package Conf;

%Config = ();		# misc top level global config data
$has_errors = 0;	# if we found errors, the web page will indicate such (global)
my $Top = undef;	# the Top pseudo-object
my @allfiles = ();	# all of the config files
my $timestamp = 0;	# time we last read the config files

# read the config file(s)
sub readconfig {
    my $file = shift;
    my $me = {};
    my( $fh, $top, $o, $nomoredata, @kids, @files );

    bless $me;
    $timestamp = $^T;
    
    $top = Group->new();
    $top->{type} = 'Group';
    $top->{name} = 'Top';

    # one file or directory of files?
    if( -d $file && -r _ && -x _ ){
	opendir D, $file;
	@files = map { "$file/$_" }
	  grep { !/^[\.\#]/ }		      # skip .file and #file
	  grep { !/(\.bkp|~)$/ } readdir D;   # and file.bkp and file~
	closedir D;
    }else{
	@files = ($file);
    }

    FILE:
    foreach $file (sort @files){
	print STDERR "reading config file: $file\n" if $::opt_d;

	push @allfiles, $file;
	$me->{fd} = $fh = BaseIO::anon_fh();
	$me->{file} = undef;
	$me->{line} = undef;
	eval {
	    open( $fh, $file ) ||
		$me->error( "'$file' is stubborn and refuses to open: $!" );
	};
	next if $@;
	$me->{file} = $file;
	$me->{line} = 0;
	eval {
	    while( defined( $_ = $me->nextline() ) ){
		print STDERR "  gotline: $_\n" if $::opt_d;
	    
		if( /^(Group|Host)\s/i ){
		    $nomoredata = 1;
		    $me->ungetline($_);
		    $o = Group::readconfig($me, $top);
		    push @kids, $o if $o;
		}
		elsif( /^Alias\s/i ){
		    $nomoredata = 1;
		    $me->ungetline($_);
		    $o = Alias::readconfig($me, $top);
		    push @kids, $o if $o;
		}
		elsif( /:/ ){
		    my ($k, $v) = split /:\s*/, $_, 2;
		    if( $nomoredata ){
			$me->warning( "additional data not permitted here (ignored)" );
			next;
		    }
		    $Config{$k} = $v;
		}
		else{
		    $me->error( "Huh? This does not look like a vaild config entry: '$_'" );
		}
	    }
	};
	if( $@ ){
	    if( $@ != $me ){
		die $@;
	    }
	    # print STDERR "caught exception ($@) $me->{file}' on line $me->{line}\n";
	    # nothing...
	}
	close $fh;
    }
    $top->{children} = [@kids];
    $top->{i_am_top} = 1;	# used only to pretty print the config
    $top->config($me);

    return if $::opt_t;
    MonEl::clearcache();        # reclaim memory used by hasattr cache
    $top->jiggle();		# update object statuses
    eval {
	$top->resolve_alii($me);
    };
    $top->{sort} = 0;     	# top is not sorted
    $top->{overridable} = 0;
    
    $top->sort_children();
    $Top = $top;
    
    $top;
}

sub error {
    my $me = shift;
    my $msg = shift;
    my( $m );

    $has_errors = 1;
    $m = "ERROR: ";
    if( $me->{file} ){
	$m .= "in file '$me->{file}' on line $me->{line} - ";
    }
    $m .= $msg;
    ::loggit( $m, 1 );

    die $me;
    undef;
}

sub warning {
    my $me = shift;
    my $msg = shift;
    my( $m );

    $m = "WARNING: ";
    if( $me->{file} ){
	$m .= "in file '$me->{file}' on line $me->{line} - ";
    }
    $m .= $msg;
    ::loggit( $m, 1 );
    undef;
}

sub nextline {
    my $me = shift;
    my( $fh, $a );

    if( defined( $a = $me->{prevline} ) ){
	delete $me->{prevline};
	return $a;
    }

    $a = '';
    $fh = $me->{fd};
    while( <$fh> ){
	$me->{line} ++;
        chop;
        s/\s*#.*$//;
        s/^\s+//;
        s/\s+$//;
        next if /^\s*$/;
        s/\\&//g;
        s/\\n/\n/g;
        s/\\r/\r/g;
        $a .= $_;
        if( $a =~ /\\$/ ){
            chop $a;
            next;
        }
        return $a;
    }

    undef;
}

sub ungetline {
    my $me = shift;
    my $line = shift;

    $me->{prevline} = $line;
}

sub check_config_files {
    foreach my $f (@allfiles){
	my $t = (stat($f))[9];
	if( $t > $timestamp ){
	    # the file changed--restart
	    ::loggit( "config file '$f' changed - restarting", 1 );
	    kill 1, $$;
	}
    }
}

# Error has no end.
#   -- Robert Browning, Paracelsus. Part iii.
sub check_typos {
    my $me = shift;
    my $obj = shift;

    foreach my $k (keys %{$obj->{config}}){
	next if $obj->{confck}{$k};
	$me->warning( "unused parameter '$k' - typo?" );
    }
}


# check for changes in config files every 5 minutes
Cron->new(
	  freq => 300,
	  text => 'Conf::check config files',
	  func => \&check_config_files,
	  );

1;

