# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <jaw @ tcp4me.com>
# Date: 2002-Oct-22 14:08 (EDT)
# Function: DNS zone authoratativeness test
#
# $Id: Domain.pm,v 1.2 2002/10/28 23:15:13 jaw Exp $

package Domain;
@ISA = qw(UDP);

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [qw(UDP Service MonEl BaseIO)],
    methods => {},
    fields => {
      dns::zone => {
	  descr => 'DNS zone to test',
	  attrs => ['config'],
      },
	
    },
};

sub probe {
    my $name = shift;

    return ( $name =~ /^UDP\/Domain/ ) ? [ 11, \&config ] : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $q, @q );
    
    bless $me;
    $me->init_from_config( $doc, 'dns' );

    if( $me->{name} =~ /Domain\/(.*)/ ){
	$me->{dns}{zone} ||= $1;
    }

    return $cf->error("zone for domain query?") unless $me->{dns}{zone};

    $q = pack( "n CC nnnn", 0, 0,0, 1,0,0,0);  # query, 1 question
    @q = split /\./, $me->{dns}{zone};
    foreach my $l (@q){
	$q .= chr(length($l)) . $l;
    }
    $q .= "\0";
    $q .= "\0\6\0\1";   # SOA IN
    
    $me->{udp}{send} = $q;
    $me->{label_right_maybe} ||= $me->{dns}{zone};

    $me->SUPER::config($cf);
    $me->{udp}{port} ||= 65;
    $me->{uname} = "Domain_$me->{dns}{zone}_$me->{srvc}{hostname}";
    
    $me;
}

sub readable {
    my $me = shift;
    my( $fh, $i, $l, $a, $b );

    $fh = $me->{fd};
    $i = recv($fh, $l, 8192, 0);
    return $me->isdown( "Domain recv failed: $!" )
	unless defined($i);

    $me->debug( "Domain recv data" );
    $me->{udp}{rbuffer} = $l;		# for debugging

    $a = vec($l, 2, 8);
    $b = vec($l, 3, 8);

    if( $b & 0xF ){
	$me->isdown( 'DNS Error (RCODE=' . ($b&0xF) . ')' );
    }
    elsif( $a & 4 ){
	$me->isup();
    }else{
	# He looked at the Gryphon as if he thought it had some kind of authority over Alice.
	#   -- Alice in Wonderland
	$me->isdown( 'DNS Error - Server Not Authoritative for Zone' );
    }
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    $me->SUPER::webpage_more($fh);
    
    foreach $k (qw(zone)){
	$v = $me->{dns}{$k};
	print $fh "<TR><TD>DNS $k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
}

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->SUPER::about_more($ctl);
    
    foreach $k (sort keys %{$me->{dns}}){
	$v = $me->{dns}{$k};
	next unless defined $v;
	if( ref($v) ){
	    $v = "#<REF>";
	}else{
	    $v = ::xxx_encode( $v );
	}
	$ctl->write( "dns::$k:\t$v\n" );
    }
}



Doc::register( $doc );
push @Service::probes, \&probe;
1;

