# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-12 20:06 (EDT)
# Function: Tell someone what happened
#
# $Id: Notify.pm,v 1.21 2002/10/25 19:59:04 jaw Exp $

package Notify;
use NotMe;
use POSIX ('_exit');

my %byid     = ();
my %lastsent = ();	# by dest
my %queue    = ();	# by dest => {not, tag}
my %unacked  = ();	# by id

# only send so often, otherwise queue
my $QUEUETIME = 120;

# we keep re-sending, but if it comes up, give up after this long
my $UNACK_TO  = 1800;

sub new {
    my $obj = shift;
    my( $me, $msg, $aa, $dst, @t );

    return unless $obj->{notify}{sendnotify};

    if( $obj->{status} eq 'down' ){
	$msg = $obj->{notify}{messagedn};
	$aa  = $obj->{notify}{autoack};
    }else{
	$msg = $obj->{notify}{messageup};
	$aa  = 1;	# up messages never need to be acked
    }

    # do not notify if user specified an empty msg
    return unless $msg;
    
    @t = localtime;
    # mm/dd hh:mm
    $msg .= " - " . sprintf "%d/%d %d:%0.2d", $t[4]+1, $t[3], $t[2], $t[1];
    $me = {
	obj      => $obj,
	msg      => $msg,
	created  => $^T,
	idno     => newid(),
	sentcnt  => 0,
	state    => 'active',		# active, supressed, acked
	objstate => $obj->{status},
	priority => $obj->{notify}{priority},
	autoack  => $aa,
	period   => $obj->{notify}{renotify},
	status   => {},
	sendto   => [],
	log      => [],

    };
    bless $me;
    return unless $me->init($obj);
    
    # RSN - supress everything else regarding this object
     
    $me->loggit( 'system', 'created' );
    $me->notify();
}

sub init {
    my $me  = shift;
    my $obj = shift;
    
    # set status of each rcpt
    $me->{sendto} = [ { start => 0, who => [] } ];
    foreach my $dst ( split /\s+/, $obj->{notify}{notify} ){
	next if $dst eq 'none';
	$me->{status}{$dst} ||= 'created';
	push @{$me->{sendto}[0]{who}}, $dst;
    }

    # He that sendeth a message by the hand of a fool cutteth off the feet, and drinketh damage.
    #   -- proverbs 26:6
    # no recipients? abort
    unless( @{$me->{sendto}[0]{who}} ){
	# QQQ this is too noisy, maybe re-enable later
	# $me->loggit( 'system', 'no recipient' );
    	return undef;
    }

    # pre-build escalation table
    # A wicked messenger falleth into mischief: but a faithful ambassador is health.
    #   -- proverbs 13:17
    my $n = 1;
    if( $obj->{notify}{escalate} ){
	foreach my $esc ( split /\;\s+/, $obj->{notify}{escalate} ){
	    # each is of form N dest dest dest ...
	    # previously, N was number of pages to escalate
	    # currently, it is elapsed-time in minutes
	    # QQQ - is there a preference?
	    my @a = split /\s+/, $esc;
	    next unless @a;
	    my $nt = shift @a;
	    $nt *= 60;  # minutes -> seconds
	    next unless $nt;
	    
	    $me->{sendto}[$n] = { start => $nt, who => [@a] };
	    $n ++;
	}
    }

    push @{$obj->{notify}{list}}, $me;
    $byid{$me->{idno}} = $me;
    
    if( $me->{state} eq 'active' ){
	$unacked{$me->{idno}} = $me;
    }

    # for easy access
    $me->{acl_ntfyack}    = $obj->{acl_ntfyack};
    $me->{acl_ntfydetail} = $obj->{acl_ntfydetail};
    
    1;
}

# flush out queued pending messages
END {
    flushqueue();
    foreach my $p (values %byid){
	$p->save();
    }
}

# return a unique id number
sub newid {
    my( $id );

    unless(open(F, "+< $::datadir/notno")){
	::loggit( "cannot open $::datadir/notno: $!", 1 );
    }
    chop( $id = <F> );
    $id ++;
    # make sure not in use
    while( exists $byid{$id} ){
	$id = int(rand(65536));
    }
    
    seek F, 0, 0;
    print F "$id\n";
    close F;

    $id;
}

sub save {
    my $me = shift;
    my $f = "$::datadir/notify/$me->{idno}";

    open( N, "> $f" ) || return $me->loggit( 'system', "cannot open save file: $!", 1 );

    print N "idno: $me->{idno}\n";
    print N "object: ", ::xxx_encode( $me->{obj}->unique() ), "\n";
    foreach my $k (qw(created msg sentcnt state objstate priority autoack
		      period ackedby ackedat escalated lastsent)){
	my $v = $me->{$k};
	next unless defined $v;

	$v = ::xxx_encode( $v );
	print N "$k: $v\n";
    }

    foreach my $dst (keys %{$me->{status}}){
	print N "status: ", ::xxx_encode($dst), " ", ::xxx_encode($me->{status}{$dst}), "\n";
    }
    
    foreach my $s (@{$me->{sendto}}){
	print N "sendto: $s->{start}";
	foreach my $dst (@{$s->{who}}){
	    print N " ", ::xxx_encode($dst);
	}
	print N "\n";
    }

    foreach my $l (@{$me->{log}}){
	print N "log: $l->{time} ", (::xxx_encode($l->{who}) || '_') , " ", ::xxx_encode($l->{msg}), "\n";
    }
    
    close N;
}

sub load {
    my $obj = shift;
    my $id  = shift;
    my $f = "$::datadir/notify/$id";
    my $me = { obj => $obj };

    open( N, $f ) || return ::loggit( "cannot open save file '$f': $!", 1 );
    while( <N> ){
	chop;

	next if( /^sendto/ );
	if( /^status:\s+([^\s]+) ([^\s]+)/ ){
	    my($dst, $st) = ($1, $2);
	    $me->{status}{ ::xxx_decode($dst) } = ::xxx_decode($st);
	}
	elsif( /^log:\s+([^\s]+) ([^\s]+) ([^\s]+)/ ){
	    my($t, $w, $m) = ($1, $2, $3);
	    push @{$me->{log}}, { time => $t,
				  who  => (($w eq '_') ? '' : ::xxx_decode($w)),
				  msg  => ::xxx_decode($m)
				  };
	}
	else{
	    my( $k, $v ) = split /:\s*/, $_, 2;
	    $me->{$k} = ::xxx_decode($v);
	}
    }

    # too old and not still active? toss
    if( $me->{state} ne 'active' && $^T - $me->{created} > (3600 * 24 * 5) ){
	unlink $f;
	return;
    }
    
    bless $me;
    $me->init($obj);
}


sub notify {
    my $me = shift;
    my $tag = shift;
    my( $n, $l, $dst, $isesc );

    # auto-ack if aa and it has already been sent
    return $me->ack() if( $me->{autoack} && $me->{sentcnt} );

    # auto-ack if in override
    return $me->ack('override') if( $me->{obj}->{ovstatus} eq 'override' );
    
    # auto-ack if up and TO
    return $me->ack('timeout') if( $me->{obj}->{status} ne 'down'
				   && $^T - $me->{obj}->{transtime} > $UNACK_TO );
    
    foreach $l (@{$me->{sendto}}){
	last if $l->{start} > $^T - $me->{created};

	if( $l->{start} ){
	    # this is an escalation
	    
	    # skip if up for X minutes;
	    if( $me->{obj}->{status} ne 'down'
		&& $^T - $me->{obj}->{transtime} > 600 ){
		$me->loggit('system', 'delaying escalation');
		last;
	    }
	    # And Saul sent messengers again the third time...
	    #   -- 1-samuel 19:21
	    $me->{escalated} = 1;
	}
	
	foreach $dst (@{$l->{who}}){
	    # send or Q
	    $me->sendorqueue($dst, $tag);
	}
    }

    $me->{sentcnt} ++;
    $me->{lastsent} = $^T;
}

sub sendorqueue {
    my $me  = shift;
    my $dst = shift;
    my $tag = shift;
    
    if( $lastsent{$dst} && ($lastsent{$dst} + $QUEUETIME > $^T) ){
	$me->queue($dst, $tag);
    }else{
	$me->transmit($dst, $tag);
    }
}

sub queue {
    my $me  = shift;
    my $dst = shift;
    my $tag = shift;	# NOT USED (yet)
    
    push @{$queue{$dst}}, {not => $me, tag => $tag};
    # print STDERR "queued on $dst - $me->{idno}\n";
    
    $me->{status}{$dst} = 'queued';
    $me->loggit( $dst, 'queued' );
    $me->save();
}

# And Hezekiah received the letter from the hand of the messengers, and read it
#   -- isaiah 37:14
sub ack {
    my $me = shift;
    my $who = shift || 'auto-ack';
    my( $dst );
    
    return undef if $me->{state} ne 'active';

    $me->{state}   = 'acked';
    $me->{ackedat} = $^T;
    $me->{ackedby} = $who;
    $me->loggit( $who, 'acked' );

    # remove from queue
    foreach $dst (keys %queue){
	@{$queue{$dst}} = grep { $_->{not}{idno} != $me->{idno} } @{$queue{$dst}};
	delete $queue{$dst} unless @{$queue{$dst}};
    }
    foreach $dst (keys %{$me->{status}}){
	$me->{status}{$dst} = "acked by $who";
    }
    
    # remove from unacked
    delete $unacked{ $me->{idno} };
    
    # stats, ...

    
    $me->save();
    
    1;
}

sub supress {
    my $me = shift;
    
    # RSN...
}

sub loggit {
    my $me = shift;
    my $who = shift;
    my $msg = shift;
    my $loudly = shift;
    
    push @{$me->{log}}, { time => $^T, who => $who, msg => $msg };
    $msg  = "<$me->{priority}> $msg" if $me->{priority};
    $msg .= " - $who" if $who;
    $me->{obj}->loggit( msg => $msg,
			tag => "NOTIFY-$me->{idno}",
			lpf => $loudly );
}

sub flushqueue {

    foreach my $dst (keys %queue){
	my @p = map { $_->{not} } @{$queue{$dst}};
	my $p = shift @p;
	next unless $p;
	$p->transmit($dst, undef, @p);
    }
}


# run by cron every minute
sub maintenance {
    my( $dst, $p, $dt, @p );

    # run queue
    foreach $dst (keys %queue){
	# print STDERR "sendq? $dst\n";
	# time to send?
	$dt = $^T - $lastsent{$dst};
	if( $lastsent{$dst} + $QUEUETIME <= $^T ){
	    # print STDERR "sending q $dst\n";
	    $p = shift @{$queue{$dst}};
	    @p = map { $_->{not} } @{$queue{$dst}};
	    # send all queued
	    $p->{not}->transmit($dst, $p->{tag}, @p) if $p;
	    delete $queue{$dst};
	}
    }

    # re-send ?
    foreach $p (values %unacked){
	$dt = $^T - $p->{lastsent};
	if( $p->{lastsent} + $p->{period} <= $^T ){
	    $p->notify( 'RESENT' );
	}
    }
}

# On this Iris, fleet as the wind, sped forth to deliver her message.
#   -- Homer, The Iliad
sub transmit {
    my $me  = shift;
    my $dst = shift;
    my $tag = shift;
    
    my @more = @_;
    my( $msg, $extra );

    my( $meth, $addr ) = split /:/, $dst;
    if( !$addr ){
	$meth = 'mail'; $addr = $dst;
    }

    # if many, summarize
    if( @more ){
	my $d;
	foreach my $p (@more){
	    $d = 1 if $p->{objstate} eq 'down';
	}
	if( $d ){
	    $msg = 'Lots of stuff just went DOWN!';
	}else{
	    $msg = 'Oodles and oodles just came UP!';
	}
    }else{
	$msg = "$me->{idno} $me->{msg}";
    }

    $extra .= " $tag" if $tag;
    $extra .= " ESCALATED" if $me->{escalated};	# RSN - from $who...

    NotMe::transmit($me, $meth, $addr, $msg, $extra)
	unless $Conf::Config{_dont_ntfy};

    $lastsent{$dst} = $^T;
    foreach my $p ( $me, @more ){
	$p->loggit( $dst, 'transmit' );
	$p->{status}{$dst} = 'sent';
	if( $p->{autoack} ){
	    $p->ack();
	}else{
	    $p->save();
	}
    }
    
}

################################################################

sub cmd_list_line {
    my $me = shift;
    my $ctl = shift;
    
    $ctl->write("$me->{idno} $me->{objstate} $me->{created} " .
		::xxx_encode($me->{obj}->filename()) . " " .
		::xxx_encode($me->{msg}) . "\n");
}

sub cmd_list {
    my $ctl = shift;
    my $param = shift;
    
    $ctl->ok();

    if( $param->{which} eq 'unacked' ){
	# oldest first
	foreach my $p (sort {$a->{created} <=> $b->{created}} values %unacked){
	    $p->cmd_list_line($ctl);
	}
    }
    elsif( $param->{which} eq 'queued' ){
	foreach my $dst (sort keys %queue){
	    # print STDERR "list q $dst\n";
	    foreach my $p (@{$queue{$dst}}){
		$p->{not}->cmd_list_line($ctl);
	    }
	}
    }
    else{
	# newest first
	foreach my $p (sort {$b->{created} <=> $a->{created}} values %byid){
	    $p->cmd_list_line($ctl);
	}
    }
    $ctl->write("\n");
}

sub cmd_detail {
    my $ctl = shift;
    my $param = shift;
    my( $p, $k, $v, $ww, $n );

    $p = $param->{idno};
    if( exists $byid{$param->{idno}} ){
	$p = $byid{$param->{idno}};
	$ctl->ok();
	foreach $k (sort keys %$p){
	    $v = $p->{$k};
	    if( ref($v) ){
		$v = "#<REF>";
	    }else{
		$v = ::xxx_encode( $v );
	    }
	    $ctl->write( "$k:\t$v\n" );
	}
	$ctl->write( "object: " . ::xxx_encode( $p->{obj}->unique() ) . "\n" );

	foreach my $w (keys %{$p->{status}}){
	    my $we = ::xxx_encode($w);
	    $ww .= " $we";
	    $ctl->write("status $we: " . $p->{status}{$w} . "\n" );
	}
	$ctl->write("statuswho: $ww\n");
	$n = 0;
	foreach my $l (@{$p->{log}}){
	    $ctl->write( "log $n: $l->{time} " . ($l->{who}? ::xxx_encode($l->{who}) : '_') .
			 ' ' . ::xxx_encode($l->{msg}) . "\n" );
	    $n++;
	}
	$ctl->write( "loglines: $n\n");
	$ctl->write( "\n" );
    }else{
	$ctl->write( "BUMMER 404 Notification Not Found\n\n" );
    }
}

sub check_acl {
    my $me = shift;
    my $g  = shift;

}

sub cmd_ack {
    my $ctl = shift;
    my $param = shift;
    my( $p, $u );
    
    $p = $param->{idno};
    $u = $param->{user} || 'anonymous';
    
    if( $p eq 'all' ){
	$ctl->ok();
	foreach $p (values %unacked){
	    $p->ack( $u );
	}
	$ctl->write("\n");
    }
    elsif( exists $byid{$p} ){
	$ctl->ok();
	$byid{$p}->ack( $u );
	$ctl->write("\n");
    }else{
	$ctl->write( "BUMMER 404 Notification Not Found\n\n" );
    }
}

################################################################

Control::command_install( 'notify_list',     \&cmd_list,
			  'list notifications', 'which' );
Control::command_install( 'notify_detail', \&cmd_detail,
			  'tell about notification', 'idno' );
Control::command_install( 'notify_ack', \&cmd_ack,
			  'ack a notification', 'idno user' );

Cron->new( freq => 60,
	   text => 'Notification maintenance',
	   func => \&maintenance,
	   );


1;
