# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-02 11:06 (EST)
# Function: misc code
#
# $Id: misc.pl,v 1.16 2002/10/28 23:15:15 jaw Exp $

# My lord the emperor, resolve me this:
#  -- Shakespeare, Titus Andronicus
# memoized gethostbyname
my %resolve = ();
sub resolve {
    my( $a ) = @_;
    my( $ip, @o );

    # skip DNS lookups if only checking the config file
    return "addr" if $::opt_t && !$::opt_T;
    
    return $resolve{$a} if( $resolve{$a} );
    
    if( $a =~ /^\d+\.\d+\.\d+\.\d+$/ ){
	$ip = inet_aton( $a );
    }else{
        $ip = gethostbyname($a);
        if( length($ip) != 4){
            sleep(1);
            $ip = gethostbyname($a);
        }
    }
    $resolve{$a} = $ip;
    $ip;
}

sub inet_aton {
    my $a = shift;

    if( $a =~ /^[\d\.]+$/ ){
	my(@a) = split /\./, $a;
	pack "C4", @a;
    }else{
	my(@a) = split /:/, $a;
	pack "n8", @a;
    }
}

sub inet_ntoa {
    my $n = shift;
    
    if( length($n) == 4 ){
	join '.', unpack( "C*", $n );
    }else{
	join ':', map {sprintf "%X",$_ } unpack( "n*", $n );
    }
}

sub ckbool {
    my $v = shift;

    ($v =~ /yes|true|on|1/i) ? 1 : 0;
}


# modification of quoted-printable encoding
# encode chars that will cause problems with webpages, etc
sub xxx_encode {
    my $txt = shift;

    $txt =~ s/([ \r\n%\+\\\;=\"\'\`\?\&~<>\/\000-\037\200-\377])/sprintf("~x%02X",ord($1))/ge;
    $txt;
}

# And so of these. Which is the natural man,
# And which the spirit? who deciphers them?
#  -- Shakespeare, Comedy of Errors
sub xxx_decode {
    my $txt = shift;

    $txt =~ s/~x(..)/chr(hex($1))/ge;
    $txt;
}

# And these does she apply for warnings, and portents,
# And evils imminent; and on her knee
#  -- Shakespeare, Julius Ceasar
sub warning {
    my $msg = shift;

    loggit( $msg, 1 );
}

sub xxx_openlog {
    if( $Conf::Config{syslog} ){
	print STDERR "opening syslog: $Conf::Config{syslog}\n" if $::opt_d;
	eval {
	    # this is only in $VER > $X
	    setlogsock( 'unix' );
	};
	openlog( $NAME, 'pid ndelay', $Conf::Config{syslog} );
    }
}

# And there was found at Achmetha, in the palace that is in the province
# of the Medes, a roll, and therein was a record thus written:
#   -- ezra 6:2
sub loggit {
    my $msg = shift;
    my $lfp = shift;
    my( $date, @d );

    # When they had thrown down their great logs of wood over the whole ground
    #   -- Homer, Iliad
    @d = localtime;
    $date = sprintf "%d/%d/%d %d:%0.2d:%0.2d",
        $d[5]+1900, $d[4]+1, $d[3], $d[2], $d[1], $d[0];

    eval {
	syslog( 'info', $msg )  if $Conf::Config{syslog};
    };
    print STDERR "[$date] $msg\n"   if $opt_f;
    
    Control::console( $msg );
    if( $lfp && $datadir ){
	if( open  LOG, ">> $datadir/log" ){
	    print LOG "[$date] [$$] $msg\n";
	    close LOG;
	}else{
	    loggit( "open log failed: $!", 0 );
	}
    }
}

sub sysproblem {
    my $msg = shift;

    warning( $msg );
}

sub trace {

    print STDERR join(", ", caller(0)), "\n";
    print STDERR join(", ", caller(1)), "\n";
    print STDERR join(", ", caller(2)), "\n";
    print STDERR join(", ", caller(3)), "\n";
    print STDERR join(", ", caller(4)), "\n";
    print STDERR join(", ", caller(5)), "\n";
    print STDERR join(", ", caller(6)), "\n";
    print STDERR "----\n";
}

sub pnow {
    print STDERR scalar(localtime $^T), "\n";
}

1;
